/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : USE_DMY_CLOCK, OUTPUT_WB_ADDR, CHECK_CONFLICT

//`define USE_DMY_CLOCK
//`define OUTPUT_WB_ADDR
`define FOWARD_DI
//`define CHECK_CONFLICT

// GPR modele for the Reduced Core TinyAVR (attiny10) 
// version 005
// 
// Target Device: xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal: Balanced
// LATCH_DOL : Insert LATCHes for Timing
//                                     USE_DMY_CLOCK    NORMAL 
// Number of Slice Flip Flops:                   52       42
// Number of 4 input LUTs:                      123      115
// Number of occupied Slices:                    78       67
// Total Number of 4 input LUTs:                130      122
// Number used for Dual Port RAMs:               16       16
// Number of bonded IOBs:                        63       63
// Post-PAR Static Timing Report
// Clock to Setup (CLK2X)                     9.496     7.139
//                (CLK)                                12.368

module rtavr_gpr_16(
		input RESET,
`ifdef USE_DMY_CLOCK
		input CLK2X_IN,
		output CLK_OUT,
`else
		input CLK2X,
		input CLK,
`endif
		input WE,
		input PREDEC, POSTINC, 
		input [7:0] DI,
		input [3:0] ADDRAL,
		input [3:0] ADDRAH,
		input [3:0] ADDRBL,
		input [3:0] ADDRBH,
`ifdef GPR_NO_DOA
`else
		output [7:0] DOAL,
		output [7:0] DOAH,
`endif
		output [7:0] DOBL,
		output [7:0] DOBH
`ifdef CHECK_CONFLICT
		, output DI_CONFLICT
`endif
		, output WB_VALID
		, output [11:0] JA
		, output [15:0] INDEX
`ifdef OUTPUT_WB_ADDR
		, output [3:0] WB_ADDR
`endif
		);
// Timing MEMO:
//
//                   _______________             ______________
// CLK2X ____________|              |____________|             |
//                                  ___________________________
// CLK  ____________________________|                          |
//
//
//      |            ------  ALL input --------                 |
//    ---- old value  ---           |  JA/INDEX low byte        |
//                                  |---|  JA       high byte   |
//                                  |----------|INDEX high byte |
//      |                      DOBL(old value)                  | DOBL
//      |                      DOBH(old value)                  | DOBH

`ifdef USE_DMY_CLOCK0
  reg [2:0] r_clk;
  wire CLK,CLK2X;
  assign CLK_OUT = r_clk[1];
  assign CLK     = r_clk[1];
  assign CLK2X   = r_clk[0];
`elsif USE_DMY_CLOCK
  reg CLK;
  wire CLK2X;
  assign CLK_OUT = CLK;
  assign CLK2X   = CLK2X_IN;
`endif
  reg  [3:0] r_addra;
  reg  [3:0] r_addrb;

  reg  [7:0] s2_DOBL;
  reg  [7:0] s2_DOBH;
  reg  [7:0] gpr [0:15];

  assign DOBL = s2_DOBL;
  assign DOBH = s2_DOBH;
`ifdef GPR_NO_DOA
`else
  reg  [7:0] s2_DOAL;
  reg  [7:0] s2_DOAH;
  assign DOAL = s2_DOAL;
  assign DOAH = s2_DOAH;
`endif
  wire [3:0] ADDRA = (CLK == 1) ? ADDRAL : ADDRAH;
  wire [3:0] ADDRB = (CLK == 1) ? ADDRBL : ADDRBH;
  wire [8:0] DOA = { 1'b0, gpr[r_addra] };

  reg  c;
  reg  r_wb_valid;
`ifdef CHECK_CONFLICT
  assign DI_CONFLICT = WE & (ADDRAL == ADDRAH);
  wire WB = (PREDEC | POSTINC) & ~DI_CONFLICT;
`else
  wire WB = (PREDEC | POSTINC);
`endif
  `ifdef OUTPUT_WB_ADDR
    reg  [3:0] r_wb_addr;
    assign WB_ADDR = r_wb_addr;
  `endif
  reg  [7:0] r_wb_data;

  wire [8:0] DOA2 =  PREDEC ? DOA-(~CLK | c) : DOA;
  wire [8:0] WB_DATA = POSTINC ? DOA+(~CLK | c): DOA2;
  
  wire WE2 = (CLK == 0) ? (WE | r_wb_valid) : WB ;
  wire [7:0] DI2 = (WE == 1) ? DI : (r_wb_valid == 1) ? r_wb_data: WB_DATA ;
  assign WB_VALID = r_wb_valid;

  reg  [7:0] s1_DOAL;
  reg  [7:0] s1_DOBL;

`ifdef FOWARD_DI
  wire FORWARD_TO_BL = (WE2) & (ADDRAH == ADDRBL);
  wire FORWARD_TO_BH = (WE2) & (ADDRAH == ADDRBH);

  wire [7:0] v_DOBL = FORWARD_TO_BL ? DI2 : s1_DOBL;
  wire [7:0] v_DOBH = FORWARD_TO_BH ? DI2 : gpr[r_addrb];
`else
  wire [7:0] v_DOBL = s1_DOBL;
  wire [7:0] v_DOBH = gpr[r_addrb];
`endif

  assign JA[7:0] = v_DOBL;
  assign JA[11:8] = v_DOBH[3:0];
  assign INDEX[7:0] = s1_DOAL;
  assign INDEX[15:8] = DOA2[7:0];

`ifdef USE_DMY_CLOCK0
always @(CLK2X_IN, RESET)
  begin
    if (RESET == 0)
      begin
	r_clk <= 0;
      end
    else
      begin
	r_clk <= r_clk + 1;
      end
  end
`elsif USE_DMY_CLOCK
always @(negedge CLK2X)
  begin
    if (RESET == 0)
      begin
	CLK <= 0;
      end
    else
      begin
	CLK <= ~CLK;
      end
  end
`endif

always @(posedge CLK2X)
  begin
    if (RESET == 0)
      begin
	c <= 0;
      end
    else
      begin
	if (WE2) gpr[r_addra] <= DI2;
	c <= WB_DATA[8];
      end
  end

always @(negedge CLK2X)
  begin
    if (RESET == 0)
      begin
	r_addra <= 0;
	r_addrb <= 0;
      end
    else
      begin
	r_addra <= ADDRA;
	r_addrb <= ADDRB;
      end
  end

always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	s1_DOAL <= 0;
	s1_DOBL <= 0;
      end
    else
      begin
	s1_DOAL <= DOA2[7:0];
	s1_DOBL <= gpr[r_addrb];
      end
  end
 
  always @(negedge CLK)
  begin
    if (RESET == 0)
      begin
`ifdef GPR_NO_DOA
`else
	s2_DOAL <= 0;
	s2_DOAH <= 0;
`endif
	s2_DOBL <= 0;
	s2_DOBH <= 0;

	r_wb_valid <= 0;
	r_wb_data <= 0;
  `ifdef OUTPUT_WB_ADDR
	r_wb_addr <= 0;
  `endif
      end
    else
      begin
`ifdef GPR_NO_DOA
`else
	s2_DOAL <= s1_DOAL;
	s2_DOAH <= DOA2[7:0];
`endif
	s2_DOBL <= v_DOBL;
	s2_DOBH <= v_DOBH;

	r_wb_valid <= WB & c;	  
	r_wb_data <= WB_DATA[7:0];
  `ifdef OUTPUT_WB_ADDR
	r_wb_addr <= ADDRAH;
  `endif
      end
  end
endmodule
